unit datum;

// un datum s'appelle aussi un Systme godsique en franais

interface

uses
  Classes,
  ellipsoide, donnee;

const
  AngleParisStr='220''14.025'''''; // angle du mridien de Paris par rapport  Greenwich

var
  AngleParis : real;

type

  TDatum = class
  private
    fEllipsoide : TEllipsoide; // elleipsoide du datum
    fDeltaX, fDeltaY, fDeltaZ : real; // dcallage du centre de l'ellipsode par
                                      // rapport au centre de la terre (WGS84) en mtres
    fReduit : boolean; // transformation rduite (3 paramtres) ou complte (7 paramtres)
    fRotX, fRotY, fRotZ : real; // rotation dcallage de l'ellipsode par
                                // rapport  WGS84 en secondes d'arc
    fScale : real; // facteur d'chelle en ppm pour passer de TRS80  l'ellipsode
    fNom : string; // nom (abrg) du datum
    fDescription : string; // description du datum
    fValid : boolean; // indique si les paramtres du datum sont bien dfinis
  public
    property Ellipsoide : TEllipsoide read fEllipsoide write fEllipsoide;
    property DeltaX : real read fDeltaX write fDeltaX;
    property DeltaY : real read fDeltaY write fDeltaY;
    property DeltaZ : real read fDeltaZ write fDeltaZ;
    property Reduit : boolean read fReduit write fReduit;
    property RotX : real read fRotX write fRotX;
    property RotY : real read fRotY write fRotY;
    property RotZ : real read fRotZ write fRotZ;
    property Scale : real read fScale write fScale;
    property Nom : string read fNom write fNom;
    property Description : string read fDescription write fDescription;
    property Valid : boolean read fValid;
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromStream (Stream : TStream; LesDonnees : TDonnees);
    procedure SaveToStream (Stream : TStream);

    // dans les deux procedures suivantes, les coordonnes cartsiennes
    // sont relatives  WGS84
    procedure VersCartesien(Lat, Long : real; Paris : boolean; var X, Y, Z : real);
    procedure VersLatLong(X, Y, Z : real; Paris : boolean; var Lat, Long : real);
    procedure Assign(Autre : TDatum);
  end;

implementation

uses
  SysUtils,
  outils, ign;

constructor TDatum.Create;
begin
  inherited Create;
  fValid:=false;
  fReduit:=true;
end;

destructor TDatum.Destroy;
begin
  inherited Destroy;
end;

procedure TDatum.LoadFromStream (Stream : TStream; LesDonnees : TDonnees);
var
  Parametres : TstringList;
  I : integer;
begin
  Parametres:=TStringList.Create;
  fValid:=false;
  try
    DetailParam(FinLigne(Stream),Parametres);
    if not (Parametres.Count in [6,10]) then
      Raise ERangeError.Create('Nombre incorrecte de paramtre lors de la lecture d''un datum');
    Nom:=Parametres[0];
    Description:=Parametres[1];
    // recherche de l'ellipsode associ
    I:=0;
    while (I<LesDonnees.LesEllipsoide.Count) and
      (TEllipsoide(LesDonnees.LesEllipsoide[I]).Nom<>Parametres[2]) do
      inc(I);
    if I>=LesDonnees.LesEllipsoide.Count then
      Raise ERangeError.CreateFmt('Ellipsode %s inconnu dans le datum %s',
      [Parametres[2],Parametres[0]]);
    Ellipsoide:=LesDonnees.LesEllipsoide[I];
    DeltaX:=StrToFloat(Parametres[3]);
    DeltaY:=StrToFloat(Parametres[4]);
    DeltaZ:=StrToFloat(Parametres[5]);
    Reduit:=(Parametres.Count=6);
    if Parametres.Count=10 then
    begin // transformation  7 paramtres
      RotX:=StrToFloat(Parametres[6]);
      RotY:=StrToFloat(Parametres[7]);
      RotZ:=StrToFloat(Parametres[8]);
      Scale:=StrToFloat(Parametres[9]);
    end;
    fValid:=true;
  finally
    Parametres.Free;
  end;
end;

procedure TDatum.SaveToStream (Stream : TStream);
var
  S : string;
begin
  S:=Format('Datum="%s" "%s" "%s" %.12g %.12g %.12g',
    [Nom, Description, Ellipsoide.Nom, DeltaX, DeltaY, DeltaZ]);
  if not Reduit then
    S:=S+Format(' %.12g %.12g %.12g %.8g',[RotX, RotY, RotZ, Scale]);
  S:=S+#13#10;
  Stream.Write(S[1],length(S));
end;

procedure TDatum.Assign(Autre : TDatum);
begin
  Nom:=Autre.Nom;
  Description:=Autre.Description;
  Ellipsoide:=Autre.Ellipsoide;
  DeltaX:=Autre.DeltaX;
  DeltaY:=Autre.DeltaY;
  DeltaZ:=Autre.DeltaZ;
  Reduit:=Autre.Reduit;
  if not Reduit then
  begin
    RotX:=Autre.RotX;
    RotY:=Autre.RotY;
    RotZ:=Autre.RotZ;
    Scale:=Autre.Scale;
  end;
end;


procedure TDatum.VersCartesien(Lat, Long : real; Paris : boolean; var X, Y, Z : real);
begin
  if Paris then
    Long:=Long+AngleParis;
  alg0009(Long, Lat, 0, Ellipsoide.a, Ellipsoide.e, X, Y, Z);
  if Reduit then
    alg0013(DeltaX, DeltaY, DeltaZ, 0, 0, 0, 0,
            X, Y, Z, X, Y, Z)
  else
    alg0013(DeltaX, DeltaY, DeltaZ, Scale/1e6,
            RotX/180/60/60*Pi, RotY/180/60/60*Pi, RotZ/180/60/60*Pi,
            X, Y, Z, X, Y, Z);
end;

procedure TDatum.VersLatLong(X, Y, Z : real; Paris : boolean; var Lat, Long : real);
var
  h : real;
begin
  if Reduit then
    alg0013bis(DeltaX, DeltaY, DeltaZ, 0, 0, 0, 0,
            X, Y, Z, X, Y, Z)
  else
    alg0013bis(DeltaX, DeltaY, DeltaZ, Scale/1e6,
            RotX/180/60/60*Pi, RotY/180/60/60*Pi, RotZ/180/60/60*Pi,
            X, Y, Z, X, Y, Z);
  alg0012(X, Y, Z, Ellipsoide.a, Ellipsoide.e, Long, Lat, h);
  if Paris then
    Long:=Long-AngleParis;
end;

begin
  DecimalSeparator:='.';
  AngleParis:=StrToAngle(AngleParisStr);
end.
